;;;; crunch - backend

(module (crunch compiler backend) (emit)

  (import (scheme base)
          (scheme cxr)
          (scheme complex)
          (scheme char)
          (chicken format)
          (chicken base)
          (chicken string)
          (chicken number-vector)
          (chicken bitwise)
          (chicken memory representation)
          (chicken plist))
  (import matchable
          srfi-1)

  (import (crunch compiler node)
          (crunch compiler type))

  (define (ctype tv #!optional box)
    (if box
        (string-append (typevar->c-type tv) " *")
        (typevar->c-type tv)))

  (define (->symbol x)
    (cond ((symbol? x) x)
          ((string? x) (string->symbol x))
          (else (error "can not convert to symbol" x))))

  (define (global-name v)
    (string-append (mangle-name (->string v)) "___"))

  (define (lambda-name id)
    (mangle-name (->string (get id 'crunch:external-name id))))

  (define (mangle-name str)
    (apply string-append
           (map (lambda (c)
                  (if (or (and (char>=? c #\A) (char<=? c #\Z))
                          (and (char>=? c #\a) (char<=? c #\z))
                          (char-numeric? c)
                          (char=? #\_ c))
                      (string c)
                      (format "_X~x" (char->integer c))))
                (string->list str))))

  (define (complex-literal? c)
    (or (string? c)
        (bytevector? c)
        (vector? c)
        (any (lambda (nc)
               (and (##core#inline "C_blockp" c) ;XXX workaround for C6 bug
                    ((car nc) c)))
             numvector-conv)))

  (define (function-signature id tv #!optional (vars '()))
    ;; returns "<rtype> <name>(<params ...>)"
    (assert (procedure-type? tv))
    (let ((extid (if (string? id)
                     id
                     (lambda-name id))))
      (if (string=? extid "main")
          (conc "int main(int " (first vars) ", char **"
                (second vars) ", char **" (third vars) ")")
          (let ((argts (procedure-argument-types tv))
                (rt (procedure-result-type tv)))
            (define (complete params)
              (string-append
                (ctype rt) " "
                extid
                "(" (string-intersperse params ",") ")"))
            (let loop ((argts argts) (params '()) (vars vars))
              (match argts
                (() (complete (reverse params)))
                (('#!optional . more)
                 (complete (cons "int argc" (append params (list "...")))))
                ((t . more)
                 (if vars
                     (loop more
                           (cons (string-append
                                   (ctype t) " "
                                   (->string (if (null? vars) (gensym 'a) (car vars))))
                             params)
                           (cdr vars))
                     (loop more (cons (ctype t) params) #f)))))))))

  (define numvector-conv
    (list (list u8vector? 'u8 u8vector-length u8vector->list 1 'uint8_t)
          (list s8vector? 's8 s8vector-length s8vector->list 1 'int8_t)
          (list u16vector? 'u16 u16vector-length u16vector->list 2 'uint16_t)
          (list s16vector? 's16 s16vector-length s16vector->list 2 'int16_t)
          (list u32vector? 'u32 u32vector-length u32vector->list 4 'uint32_t)
          (list s32vector? 's32 s32vector-length s32vector->list 4 'int32_t)
          (list u64vector? 'u64 u64vector-length u64vector->list 8 'uint64_t)
          (list s64vector? 's64 s64vector-length s64vector->list 8 'int64_t)
          (list f32vector? 'f32 f32vector-length f32vector->list 4 'float)
          (list f64vector? 'f64 f64vector-length f64vector->list 8 'double)
          (list c64vector? 'c64 c64vector-length c64vector->list 8 "float complex")
          (list c128vector? 'c128 c128vector-length c128vector->list 16 "double complex")))

  (define (vector->bitlist v)
    (let ((len (vector-length v)))
      (let loop ((i 7) (lst '()) (p 0) (n 0))
        (if (>= p len)
            (reverse lst)
            (if (zero? i)
                (loop 7 (cons n lst) (add1 p) 0)
                (loop (sub1 i) lst (add1 p)
                      (bitwise-ior (arithmetic-shift n 1)
                                   (if (vector-ref v p) 1 0))))))))

  (define (emit n consumer)
    (let ((deferred '())
          (literals '())
          (procs '()))
      (define (out . args) (consumer (apply conc args)))
      (define (outnums nums)
        (for-each (lambda (n) (out n ",")) nums))
      (define (literal id c tv)
          (cond ((string? c)
                 (let ((bs (u8vector->list (string->utf8 c))))
                   (out "static " (ctype tv) "_block " id " = { -1, ")
                   (out "0, " (string-length c) ", " (add1 (length bs)) ", \"")
                   (for-each
                     (lambda (b)
                       (let ((n (number->string b 8)))
                         (out "\\" (make-string (- 3 (string-length n)) #\0) n)))
                     bs)
                   (out "\" };\n")
                   (conc "&" id)))
                ((and (procedure? c)
                	  (extended-procedure? c)
                	  (match (procedure-data c)
                	    (('crunch:lambda n) n)
                	    (_ #f))) =>
                  (lambda (n)
                    (conc "(void *)" (first (node-data n)))))
                ((vector? c)
                 (case (typevar-type (vector-element-type tv))
                   ((boolean)
                    (let* ((len (vector-length c))
                           (blen (quotient (+ len 7) 8))
                           (id2 (gensym 'b)))
                      (out "static uint8_t " id2 "[] = { ")
                      (outnums (vector->bitlist c))
                      (out " };\nstatic crunch_u8vector_block " id " = { -1, ")
                      (out "0, " len ", " blen ", " id2 " };\n")
                      (conc "&" id)))
                   ((char)
                    (let* ((len (vector-length c))
                           (id2 (gensym 'b)))
                      (out "static uint32_t " id2 "[] = { ")
                      (outnums (map char->integer (vector->list c)))
                      (out " };\nstatic crunch_u32vector_block " id " = { -1, ")
                      (out "0, " len ", " (* 4 len) ", " id2 " };\n")
                      (conc "&" id)))
                   ((integer)
                    (let ((len (vector-length c))
                          (id2 (gensym 'b)))
                      (out "static crunch_word " id2 "[] = { ")
                      (outnums (vector->list c))
                      (out " };\nstatic " (ctype tv) "_block " id " = { -1, ")
                      (out "0, " len ", " len " * sizeof(crunch_word) , " id2
                           " };\n")
                      (conc "&" id)))
                   ((float)
                    (literal id (list->f64vector (vector->list c)) #f))
                   ((complex)
                    (literal id (list->c128vector (vector->list c)) #f))
                   (else
                     (let* ((lst (vector->list c))
                            (etv (vector-element-type tv))
                            (len (vector-length c))
                            (idb (gensym 'ab))
                            (lits (map (lambda (x)
                                         (literal (gensym 'lb) x etv))
                                       lst)))
                       (out "static struct crunch_block *" idb "[ " len " ] = { ")
                       (for-each (cut out "(struct crunch_block *)" <> ", ") lits)
                       (out " };\nstatic crunch_vector_block " id " = { -1, ")
                       (out "CRUNCH_NESTEDRC, " len ", " len " * sizeof(crunch_word), "
                            idb " };\n")
                       (conc "&" id)))))
                ((number-vector? c)
                 (let loop ((nc numvector-conv))
                   (if ((first (car nc)) c)
                       (let* ((conv (car nc))
                              (nm (second conv))
                              (len ((third conv) c)))
                         (cond ((positive? len)
                                (let ((id2 (gensym 'vb))
                                      (nb (gensym 'nb))
                                      (mul (fifth conv)))
                                  (out "static " (sixth conv) " " nb "[] = {")
                                  (outnums ((fourth conv) c))
                                  (out "};\nstatic crunch_" nm "vector_block "
                                       id2 " = { -1, 0, " len ", " (* len mul)
                                       ", " nb " };\n")
                                  (conc "&" id2)))
                               (else
                                 (out "#define " id " ((crunch_" nm
                                      "vector)&crunch_null_vector)\n")
                                 id)))
                       (loop (cdr nc)))))
                (else (error 'crunch "backend: can not encode literal" c))))
      (define (default-value tv)
        (match (typevar-type tv)
          ('string (string-append "(" (ctype tv) ")&crunch_null_string"))
          (((or 'numvector 'vector) _)
            (string-append "(" (ctype tv) ")&crunch_null_vector"))
          ('port (string-append "(" (ctype tv) ")&crunch_null_port"))
          (else "0")))
      ;; prototypes and includes
      (define (protos n exp?)
        (case (node-class n)
          ((const)
            (let ((c (first (node-data n)))
                  (tv (node-type n)))
              (when (complex-literal? c)
                (cond ((assq c literals) =>
                        (lambda (a) (set-cdr! (node-data n) (list (cdr a)))))
                      (else
                        (let ((id (gensym 'c))
                              (lit (literal (gensym 'lb) c (node-type n))))
                          (out "static " (ctype tv) " " id " = "
                               (if (eq? (typevar-type tv) 'port)
                                   "(crunch_port)" "")
                               lit ";\n")
                          (set-cdr! (node-data n) (list id))
                          (set! literals (cons (cons c id) literals))))))))
          ((global-set!)
           (match (node-data n)
             ((_ exp? _)
              (protos (first (node-subnodes n)) #t))))
          ((lambda)
           (match (node-data n)
             ((id argc _ as _)
               (let ((sig (function-signature id (node-type n) as)))
                 (out (if exp? "" "static ") sig ";\n")
                 (set! procs (cons (cons n exp?) procs))
                 (protos (first (node-subnodes n)) #f)))))
          ((c-external)
           (and-let* ((inc (first (node-data n))))
             (if (string? inc)
                 (out "#include \"" inc "\"\n")
                 (out "#include " inc "\n"))))
          ((c-lambda)
           (match (node-data n)
             ((name ret args body)
               (let ((extid (get name 'crunch:external-name)))
                 (out "\n" (if extid "" "static ") (ctype (typespec ret)) " "
                      (if extid (lambda-name extid) name) "("))
               (when (pair? args)
                 (out (ctype (typespec (caar args))) " " (cadar args))
                 (for-each
                   (lambda (arg)
                     (out ", " (ctype (typespec (car arg))) " " (cadr arg)))
                   (cdr args)))
               (out ") {\n")
               (for-each (cut out <> "\n") body)
               (out "}\n\n"))))
          ((idcall)
            (let ((id (first (node-data n))))
              ;; extern prototype if imported
              (when (and (symbol? id)
                         (not (get id 'crunch:lambda-node))
                         (not (get id 'crunch:primitive)))
                (and-let* ((t (get id 'crunch:extern)))
                  (put! id 'crunch:extern #f) ; so only once generated
                  (out "extern " (function-signature id t #f) ";\n")))
              (for-each (cut protos <> #f) (node-subnodes n))))
          ((c-declare)
           (for-each (cut out <> "\n") (first (node-data n))))
          (else (for-each (cut protos <> #f) (node-subnodes n)))))
      ;; toplevel expressions
      (define (top n exp?)
        (case (node-class n)
          ((begin)
           (for-each (cut top <> #f) (node-subnodes n)))
          ((global-set!)
           (match (node-data n)
             ((var exp? lvar)
              (let ((val (first (node-subnodes n))))
                (top val exp?)
                (let ((tv (node-type val)))
                  (cond ((procedure-type? tv)
                         (out (if exp? "" "static ")
                              (ctype tv) " " (global-name lvar)
                              " = "
                              (match (pure-value-set tv)
                                (((? symbol? id)) (lambda-name id))
                                ;;XXX ?
                                (_ (or (get var 'crunch:initial-value)
                                       (default-value tv))))
                              ";\n"))
                        ((eq? 'const (node-class val))
                         (match (node-data val)
                           ((_ id)
                            ;;XXX ?
                            (out "#define " (global-name lvar) " " id "\n"))
                           ((c)
                            (out (if exp? "" "static ")
                                 (ctype tv) " " (global-name lvar)
                                 " = "
                                 (or (get var 'crunch:initial-value) c)
                                 ";\n"))))
                        (else
                          (out (if exp? "" "static ")
                               (ctype tv) " " (global-name lvar)
                               " = "
                               (if (eq? (typevar-type tv) 'port)
                                   "(crunch_port)&crunch_null_port"
                                   (default-value tv))
                               ";\n"))))))))
          ((variant) (for-each (cut top <> #f) (node-subnodes n)))
          ((const unspecified c-lambda c-external c-declare lambda) #f)
          ;; all others
          (else
            => (lambda (c)
                 (error "backend: invalid toplevel node class" c)))))
      ;; procedure declaration and body
      (define (procedure n exp?)
        (match (node-data n)
          ((id argc _ as _)
           (let* ((tv (node-type n))
                  (body (first (node-subnodes n)))
                  (main? (and exp? (string=? "main" (lambda-name id))))
                  (sig (function-signature id tv as)))
             (set! deferred '())
             (out "\n" (if exp? "" "static ")
                  sig
                  " {\n")
             (let loop ((argts (procedure-argument-types tv))
                        (as as)
                        (last "argc"))
               (match argts
                 (('#!optional more ...)
                    (out "va_list va;\nva_start(va, " last ");\n")
                    (let loop2 ((argts more) (as as) (i argc))
                      (unless (null? argts)
                        (let* ((a (car as))
                               (box? (get a 'crunch:boxed))
                               (ba (if box? (gensym 'b) (car as)))
                               (ct (ctype (car argts))))
                          (out ct " " ba ";\n")
                          (out "if(argc > " i ") " ba " = va_arg(va, " ct ");\n")
                          (when box?
                            (out (ctype (car argts) #t) " " a " = &" ba ";}\n"))
                          (loop2 (cdr argts) (cdr as) (add1 i))))))
                 ((tv more ...)
                  (loop more (cdr as) (car as)))
                 (() #f)))
             (when main?
               (out "crunch_main(" (car as) ", " (cadr as) ", " (caddr as) ");\nlong r;\n")
               (set! deferred '(#t))
               (when (not (numeric-type? (node-type body)))
                 (set! body (make-node (node-location body)
                                       'begin '()
                                       (list body (make-node (node-location n)
                                                             'const (list 0) '()
                                                             (typespec 'integer)))
                                       (typespec 'integer)))))
             (return body #t)
             (when (pair? deferred)
               (out "rl:\n")
               (for-each
                 (lambda (d)
                   (if (eq? d #t) ; main
                       (out "crunch_cleanup();\n")
                       (out "crunch_unref(" (if (cdr d) "*" "") (car d) ");\n")))
                 deferred)
               (return-value (node-type body) main?))
             (out "}\n")))))
      (define (return-value tv main?)
        (cond ((and main? (not (numeric-type? tv)))
                (out "return 0;\n"))
              ((eq? 'void (typevar-type tv))
                (out "return;\n"))
              ((refcounted-type? tv)
                (out "return crunch_unref_result(r);\n"))
              (else (out "return r;\n"))))
      (define (declare-result tv)
        (out (ctype tv) " r")
        (when (refcounted-type? tv)
          (out " = (" (ctype tv) ")&crunch_null_vector"))
        (out ";\n"))
      ;; statements
      (define (return n tail?)
        (define (voidreturn)
          (if (pair? deferred)
              (out "goto rl;\n")
              (out "return;\n")))
        (define (valreturn n)
          (cond ((pair? deferred)
                  (let ((tv (node-type n)))
                    (cond ((refcounted-type? tv)
                            (out "crunch_move(" (ctype tv) ", ")
                            (expr n #f)
                            (out ", r);\n"))
                          (else
                            (out "r = ")
                            (expr n #f)
                            (out ";\n")))
                    (out "goto rl;\n")))
                (else
                  (out "return(")
                  (expr n #f)
                  (out ");\n"))))
        (if tail?
            (case (node-class n)
              ((unspecified c-lambda c-declare) (voidreturn))
              ((const global-ref local-ref lambda box-ref c-external idcall call)
                (case (typevar-type (node-type n))
                  ((void)
                    (expr n #f)
                    (out ";\n")
                    (voidreturn))
                  ((noreturn)
                    (expr n #f)
                    (out ";\n"))
                  (else (valreturn n))))
              ;; let, labels, goto, if, begin, dispatch
              (else (stmt n #t)))
            (stmt n #f)))
      (define (stmt n tail?)
        (and-let* ((loc (node-location n)))
          (out "// " loc "\n"))
        (case (node-class n)
          ((if)
           (match (node-subnodes n)
             ((tn cn an)
              (out "if(")
              (expr tn #f)
              (out ") {\n")
              (return cn tail?)
              (out "} else {\n")
              (return an tail?)
              (out "}\n"))))
          ((switch)
            (match (node-subnodes n)
              ((val yes no)
               (out "if(")
               (expr val #f)
               (out " == " (first (node-data n)) ") {\n")
               (return yes tail?)
               (out "} else {\n")
               (return no tail?)
               (out "}\n"))
              ((val cases ...)
               (out "switch(")
               (expr val #f)
               (out ") {\n")
               (let ((sl (gensym 'sl)))
                 (for-each
                   (lambda (c n)
                     (out "case " c ":\n")
                     (return n tail?)
                     (out "goto " sl ";\n"))
                   (first (node-data n))
                   (butlast cases))
                 (out "default:\n")
                 (return (last cases) tail?)
                 (out "}\n" sl ":;\n")))))
          ((global-set!)
           (assign (first (node-subnodes n))
                   (lambda () (out (global-name (third (node-data n)))))
                   #t)
           (out ";\n"))
          ((local-set!)
           (assign (first (node-subnodes n))
                   (lambda () (out (first (node-data n))))
                   #f)
           (out ";\n"))
          ((box-set!)
           (assign (second (node-subnodes n))
                   (lambda ()
                     (out "*")
                     (expr (first (node-subnodes n)) #f))
                   #f)
           (out ";\n"))
          ((begin)
           (for-each (cut stmt <> #f) (butlast (node-subnodes n)))
           (return (last (node-subnodes n)) tail?))
          ((call idcall)
           (expr n tail?)
           (out ";\n"))
          ((dispatch)
           (let* ((var (gensym 't))
                  (ns (node-subnodes n))
                  (fn (first ns)))
             (out "crunch_function " var " = ")
             (expr fn #f)
             (out ";\n")
             (let rec ((ns (cdr ns)) (ids (first (node-data n))))
               (match ns
                 ((call more ...)
                  (let ((id (car ids)))
                    (out "if(" (lambda-name id) " == " var ") {\n")
                    (return call tail?)
                    (out "}\n")
                    (when (pair? more)
                      (out "else {\n")
                      (rec (cdr ns) (cdr ids))
                      (out "}\n"))))))))
          ((goto)
           (out "goto " (first (node-data n)) ";\n"))
          ((lambda) #f)
          ((let)
           (let ((as (second (node-data n)))
                 (ns (node-subnodes n)))
             (for-each
               (lambda (a vn)
                 (let* ((tv (node-type vn))
                        (box? (get a 'crunch:boxed))
                        (ba (if box? (gensym 'b) a))
                        (d (refcounted-type? tv)))
                   (when d
                     (let ((ntv (node-type n)))
                       (when (and (null? deferred)
                                  (not (eq? 'void (typevar-type ntv))))
                         (declare-result ntv)))
                     (set! deferred (cons (cons a box?) deferred)))
                   (out (ctype tv) " " ba)
                   (unless (uninitialized-type? tv)
                     (out " = ")
                     (cond ((and (eq? 'const (node-class vn))
                                 (equal? '#() (first (node-data vn))))
                             ;; arbitrary vector, needs cast
                             (out "(" (ctype tv) ")&crunch_null_vector"))
                           ((and (eq? 'port (typevar-type tv))
                                 (eq? 'const (node-class vn))) ; default init val
                             (out "(crunch_port)&crunch_null_port"))
                           (else (expr vn #f))))
                   (out ";\n")
                   (when box?
                     (out (ctype tv box?) " " a " = &" ba ";\n"))
                   (when d
                     (out "crunch_addref("
                          (if box? "*" "")
                          a ");\n"))))
               as (butlast ns))
             (return (last ns) tail?)))
          ((labels)
           (for-each
             (lambda (l n)
               (out l ":\n")
               (return n tail?))
             (first (node-data n))
             (node-subnodes n)))
          (else
            (expr n tail?)
            (out ";\n"))))
      (define (assign val dest pin)
        (let ((tv (node-type val)))
          (cond ((refcounted-type? tv)
                  (out "crunch_move" (if pin "_and_pin" "") "(" (ctype tv) ", ")
                  (expr val #f)
                  (out ", ")
                  (dest)
                  (out ")"))
                (else
                  (dest)
                  (out " = ")
                  (expr val #f)))))
      ;; inner expressions
      (define (expr n tail?)
        (case (node-class n)
          ((const)
            (match (node-data n)
              ((_ id) (out id))
              ((c)
                (cond ((char? c) (out (char->integer c)))
                      ((boolean? c) (out (if c 1 0)))
                      ((and (number? c) (not (real? c)))
                        (out "(" (real-part c) " + I * " (imag-part c) ")"))
                      (else (out c))))))
          ((unspecified) #f)
          ((global-ref) (out (global-name (third (node-data n)))))
          ((local-ref) (out (first (node-data n))))
          ((if)
           (match (node-subnodes n)
             ((tn cn an)
              (out "(")
              (expr tn #f)
              (out " ? ")
              (expr cn tail?)
              (out " : ")
              (expr an tail?)
              (out ")"))))
          ((switch)
            (let ((val (first (node-subnodes n))))
              (let rec ((xs (first (node-data n)))
                        (ns (cdr (node-subnodes n))))
                (if (null? xs)
                    (expr (car ns) tail?)
                    (begin
                      (out "(")
                      (expr val #f)
                      (out " == " (car xs) " ? ")
                      (expr (car ns) tail?)
                      (out " : ")
                      (rec (cdr xs) (cdr ns))
                      (out ")"))))))
          ((begin)
           (out "(")
           (for-each (lambda (n)
                       (expr n #f)
                       (out " ,"))
             (butlast (node-subnodes n)))
           (expr (last (node-subnodes n)) tail?)
           (out ")"))
          ((global-set!)
           (out "(")
           (assign (first (node-subnodes n))
                   (lambda () (out (global-name (third (node-data n)))))
                   #t)
           (out ")"))
          ((local-set!)
           (out "(")
           (assign (first (node-subnodes n))
                   (lambda () (out (first (node-data n))))
                   #f)
           (out ")"))
          ((box-set!)
           (out "(")
           (assign (second (node-subnodes n))
                   (lambda ()
                     (out "*")
                     (expr (first (node-subnodes n)) #f))
                   #f)
           (out ")"))
          ((lambda)
           (let ((id (first (node-data n))))
             (out (lambda-name id))))
          ((call)
           (let* ((ns (node-subnodes n))
                  (fn (first ns))
                  (fnt (node-type fn))
                  (opt (memq '#!optional (procedure-argument-types fnt))))
             (out "((" (function-signature "(*)" fnt #f) ")(")
             (when opt
               (out (length (node-subnodes n))
                    (if (pair? ns) ", " "")))
             (for-each (lambda (n)
                         (out ", ")
                         (expr n #f))
               ns)
             (out ")")))
          ((idcall)
           (let* ((id (first (node-data n)))
                  (ns (node-subnodes n)))
             (match id
               ((? symbol?)
                (out (lambda-name id) "(")
                (when (third (node-data n))
                  (out (length ns))
                  (unless (null? ns) (out ", ")))
                (unless (null? ns)
                  (expr (first ns) #f)
                  (for-each
                    (lambda (n)
                      (out ", ")
                      (expr n #f))
                    (cdr ns)))
                (out ")"))
               (('operator op)
                (out "(")
                (cond ((null? (cdr ns))
                       (out op)
                       (expr (first ns) #f))
                      (else
                        (expr (first ns) #f)
                        (out " " op " ")
                        (expr (second ns) #f)))
                (out ")"))
               (('operator op val)
                (out "(")
                (expr (first ns) #f)
                (out " " op " " val ")")))))
          ((dispatch)
           (let* ((var (gensym 't))
                  (r (gensym 't))
                  (ns (node-subnodes n))
                  (fn (first ns)))
             (out "({ " (ctype (node-type n)) " " r "; "
                  "crunch_function " var " = ")
             (expr fn #f)
             (out "; ")
             (let rec ((ns (cdr ns)) (ids (first (node-data n))))
               (match ns
                 ((call more ...)
                  (let loop ((id (car ids)))
                    (out "if(" (lambda-name id) " == " var ") { "
                         r " = ")
                    (expr call tail?)
                    (out "; } ")
                    (when (pair? more)
                      (out "else {")
                      (rec more (cdr ids))
                      (out "} "))))))
             (out r ";})")))
          ((box-ref)
           (out "*")
           (expr (first (node-subnodes n)) #f))
          ((wrap)
           (out (first (node-data n)) "(")
           (expr (first (node-subnodes n)) #f)
           (out ")"))
          ((c-external)
           (out (second (node-data n))))
          ;; variant, goto, let, labels
          (else => (lambda (c) (error "backend: invalid expression node class" c)))))
      (out "/* GENERATED BY CRUNCH */\n\n"
           "#ifndef CRUNCH_H\n#include \"crunch.h\"\n#endif\n\n")
      (protos n #f)
      (out "\n")
      (top n #f)
      (for-each (lambda (nx) (procedure (car nx) (cdr nx))) procs)
      (out "/* END OF GENERATED CODE */\n")
      (consumer #!eof)))

)
