;;; CRUNCH - debugger for node trees

(module (crunch compiler debug)
  (emit-debug-message debug-consumer register-pseudo-source-file
   debug-node-tree debugger-active connect-to-debugger
   register-source-file)

(import scheme (scheme base))
(import chicken.base chicken.fixnum chicken.keyword chicken.string)
(import srfi-1)
(import (chicken format) (chicken irregex)
        (chicken process-context) (chicken platform)
        (chicken memory representation)
        (chicken plist)
        (chicken process) (chicken string))
(import (crunch compiler node)
        (crunch compiler type))

(define debugger-active (make-parameter #f))

(define wish (or (get-environment-variable "CRUNCH_WISH") "wish"))

(define crunch-dir
  (or (get-environment-variable "CRUNCH_DIR")
      (chicken-home)))

(define debugger-input (make-parameter #f))

; Adapted from Scheme pretty-printer by Marc Feeley:
;
; Copyright (c) 1991, Marc Feeley
; Author: Marc Feeley (feeley@iro.umontreal.ca)
; Distribution restrictions: none
;
; Modified by felix for use with CHICKEN, then CRUNCH

(define generic-write
  (lambda (obj display? width output)

    (define (out str col)
      (and col (output str) (+ col (string-length str))))

    (define (output-id-list lst)
      (if (null? lst)
          (output "(none)")
          (for-each (lambda (x) (output (format "<~a> " x))) lst)))

    (define (clean-vs vs)
      (map (lambda (v)
      	     (if (string? v)
      	     	 (let ((v (string-translate* v
      	     	            '(("\n" . "\\\\n") ("\\" . "\\\\") ("\"" . "\\\"")))))
      	     	   (if (> (string-length v) 10)
      	     	       (string-append (substring v 0 10) "...")
      	     	       v))
      	     	 v))
      	   vs))

    (define (begin-tag n)
      (let* ((tv (node-type n))
      	     (nid (gensym 'n))
      	     (tid (typevar-id tv))
      	     (ts (typename tv))
      	     (vs (typevar-value-set tv)))
        (output (format "\"\nbtags ~a t~a \"~a\" \"~a\"" nid tid ts
                  (clean-vs vs))))
      (and-let* ((loc (node-location n)))
        (when (and (string? loc)
        	   (irregex-search '(: bos (+ (~ #\:)) #\:) loc))
          (output (format " \"~a\"" loc))))
      (output "\n")
      (when (eq? (node-class n) 'lambda)
        (let ((id (first (node-data n))))
          (output (format "ftag <~a> \"\\ncallers: " id))
          (output-id-list (get id 'crunch:callers '()))
          (output "\\ncallees: ")
          (output-id-list (get id 'crunch:callees '()))
          (and-let* ((n (get id 'crunch:external-name)))
            (output (format "\\nexternal-name: ~a" n)))
          (when (get id 'crunch:escapes) (output "\\nescapes"))
          (when (get id 'crunch:lambda-literal) (output "\\nused as literal"))
	  (output "\"\n")))
      (output "txt \""))

    (define (end-tag)
      (output "\"\netags\ntxt \""))

    (define (wr obj col)

      (define (wr-lst l col)
        (if (pair? l)
            (let loop ((l (cdr l))
                       (col (and col (wr (car l) (out "\x28;" col)))))
              (cond ((not col) col)
                    ((pair? l)
                     (loop (cdr l) (wr (car l) (out " " col))))
                    ((null? l) (out "\x29;" col))
                    (else      (out "\x29;" (wr l (out " . " col))))))
            (out "()" col)))

      (cond ((and (pair? obj) (eq? 'quote (car obj)))
              (wr (cadr obj) (out "'" col)))
            ((pair? obj)        (wr-lst obj col))
            ((null? obj)        (wr-lst obj col))
            ((vector? obj)      (wr-lst (vector->list obj) (out "#" col)))
            ((boolean? obj)     (out (if obj "#t" "#f") col))
            ((number? obj)      (out (number->string obj) col))
            ((or (keyword? obj) (symbol? obj))
             (let ((s (open-output-string)))
               (##sys#print obj #t s)
               (out (get-output-string s) col) ) )
            ((node? obj)
              (begin-tag obj)
              (let ((r (wr (node->expr obj) col)))
                (end-tag)
                r))
            ((string? obj)
             (if display?
                 (out obj col)
                 (let loop ((i 0) (j 0) (col (out "\\\"" (sub1 col))))
                   (if (and col (fx< j (string-length obj)))
                       (let ((c (string-ref obj j)))
                         (cond
                          ((or (char=? c #\\)
                               (char=? c #\"))
                           (loop j
                                 (+ j 1)
                                 (out "\\"
                                      (out (substring obj i j) col))))
                          ((or (char<? c #\x20)
                               (char>=? c #\x7f))
                           (loop (+ j 1)
                                 (+ j 1)
                                 (let ((col2 (out (substring obj i j) col)))
                                   (cond ((assq c '((#\tab . "\\t")
                                                    (#\newline . "\\n")
                                                    (#\return . "\\r")))
                                          =>
                                          (lambda (a)
                                            (out (cdr a) col2)))
                                         (else
                                          (out (string-append
                                                 "\\x"
                                                 (number->string (char->integer c) 16)
                                                 ";")
                                               col2))))))
                          (else (loop i (+ j 1) col))))
                       (out "\\\""
                            (out (substring obj i j) (sub1 col)))))))
            ((char? obj)        (if display?
                                    (out (string obj) col)
                                    (let ((code (char->integer obj))
                                          (col2 (out "#\\\\" (sub1 col))))
                                      (cond ((char-name obj)
                                             => (lambda (cn)
                                                  (out (symbol->string cn) col2) ) )
                                            ((or (fx< code 32) (fx> code 127))
                                             (out
                                               ";"
                                               (out (number->string code 16) col2)))
                                            (else (out (string obj) col2)) ) ) ) )
            ((##core#inline "C_bytevectorp" obj)
             (wr-lst (##sys#bytevector->list obj) (out "#u8" col)))
            (else (out "#<unprintable object>" col)) ) )

    (define (pp obj col)

      (define (spaces n col)
        (if (> n 0)
            (if (> n 7)
                (spaces (- n 8) (out "        " col))
                (out (substring "        " 0 n) col))
            col))

      (define (indent to col)
        (and col
             (if (< to col)
                 (and (out "\\n" col) (spaces to 0))
                 (spaces (- to col) col))))

      (define (pr obj col extra pp-pair)
        (cond ((node? obj)
               (begin-tag obj)
               (let ((r (pr (node->expr obj) col extra pp-pair)))
                 (end-tag)
                 r))
              ((and (pair? obj) (eq? 'quote (car obj)))
                (pr (cadr obj) (out "'" col) extra pp-expr))
              ((or (pair? obj) (vector? obj)) ; may have to split on multiple lines
               (let ((result '())
                     (left (max (+ (- (- width col) extra) 1) max-expr-width)))
                 (generic-write obj display? #f
                                (lambda (str)
                                  (set! result (cons str result))
                                  (set! left (- left (string-length str)))
                                  (> left 0)))
                 (if (> left 0)          ; all can be printed on one line
                     (out (reverse-string-append result) col)
                     (if (pair? obj)
                         (pp-pair obj col extra)
                         (pp-list (vector->list obj) (out "#" col) extra pp-expr)))))
              (else (wr obj col))))

      (define (pp-expr expr col extra)
        (if (node? expr)
            (let ((proc (style expr)))
              (if proc
                  (proc expr col extra)
                  (let* ((x (node->expr expr))
                         (head (car x)))
                    (begin-tag expr)
                    (let ((r (if (> (string-length (symbol->string head))
                                    max-call-head-width)
                                 (pp-general x col extra #f #f #f pp-expr)
                                 (pp-call x col extra pp-expr))))
                      (end-tag)
                      r))))
            (pp-list expr col extra pp-expr)))

                                        ; (head item1
                                        ;       item2
                                        ;       item3)
      (define (pp-call expr col extra pp-item)
        (let ((col* (wr (car expr) (out "\x28;" col))))
          (and col
               (pp-down (cdr expr) col* (+ col* 1) extra pp-item))))

                                        ; (item1
                                        ;  item2
                                        ;  item3)
      (define (pp-list l col extra pp-item)
        (let ((col (out "\x28;" col)))
          (pp-down l col col extra pp-item)))

      (define (pp-down l col1 col2 extra pp-item)
        (let loop ((l l) (col col1))
          (and col
               (cond ((pair? l)
                      (let ((rest (cdr l)))
                        (let ((extra (if (null? rest) (+ extra 1) 0)))
                          (loop rest
                                (pr (car l) (indent col2 col) extra pp-item)))))
                     ((null? l)
                      (out "\x29;" col))
                     (else
                      (out "\x29;"
                           (pr l
                               (indent col2 (out "." (indent col2 col)))
                               (+ extra 1)
                               pp-item)))))))

      (define (pp-general expr col extra named? pp-1 pp-2 pp-3)

        (define (tail1 rest col1 col2 col3)
          (if (and pp-1 (pair? rest))
              (let* ((val1 (car rest))
                     (rest (cdr rest))
                     (extra (if (null? rest) (+ extra 1) 0)))
                (tail2 rest col1 (pr val1 (indent col3 col2) extra pp-1) col3))
              (tail2 rest col1 col2 col3)))

        (define (tail2 rest col1 col2 col3)
          (if (and pp-2 (pair? rest))
              (let* ((val1 (car rest))
                     (rest (cdr rest))
                     (extra (if (null? rest) (+ extra 1) 0)))
                (tail3 rest col1 (pr val1 (indent col3 col2) extra pp-2)))
              (tail3 rest col1 col2)))

        (define (tail3 rest col1 col2)
          (pp-down rest col2 col1 extra pp-3))

        (let* ((head (car expr))
               (rest (cdr expr))
               (col* (wr head (out "\x28;" col))))
          (if (and named? (pair? rest))
              (let* ((name (car rest))
                     (rest (cdr rest))
                     (col** (wr name (out " " col*))))
                (tail1 rest (+ col indent-general) col** (+ col** 1)))
              (tail1 rest (+ col indent-general) col* (+ col* 1)))))

      (define (pp-expr-list l col extra)
        (pp-list l col extra pp-expr))

      (define (pp-lambda expr col extra)
        (pp-general expr col extra #f pp-expr-list #f pp-expr))

      (define (pp-if expr col extra)
        (pp-general expr col extra #f pp-expr #f pp-expr))

      (define (pp-let expr col extra)
        (let* ((rest (cdr expr))
               (named? (and (pair? rest) (symbol? (car rest)))))
          (pp-general expr col extra named? pp-expr-list #f pp-expr)))

      (define (pp-begin expr col extra)
        (pp-general expr col extra #f #f #f pp-expr))

      (define (pp-const expr col extra)
	(pr (node->expr expr) col extra pp-expr))

      ;; define formatting style (change these to suit your style)

      (define indent-general 2)

      (define max-call-head-width 5)

      (define max-expr-width 50)

      (define (style n)
        (case (node-class n)
          ((const)                     pp-const)
          ((lambda) 		       pp-lambda)
          ((if global-set! local-set!) pp-if)
          ((let)                       pp-let)
          ((begin)                     pp-begin)
          (else                        #f)))

      (pr obj col 0 pp-expr))

    (if width
        (out "\\n" (pp obj 0))
        (wr obj 0))))

(define (escape str)
  (string-translate* str '(("[" . "\\[") ("]" . "\\]") ("$" . "\\$"))))

(define (pretty-print obj width port)
  (display "txt \"" port)
  (generic-write obj #f width (lambda (s) (display (escape s) port) #t))
  (display "\"\n" port))

(define (debug-node-tree n title)
  (display (escape (format "title ~s\n" title)) (debugger-input))
  (pretty-print n 200 #;(current-error-port) (debugger-input)))

(define (emit-debug-message str)
  (display (escape (format "msg ~s\n" str)) (debugger-input)))

(define (debug-consumer)
  (fprintf (debugger-input) "debugging\n")
  void)

(define (connect-to-debugger k)
  (receive (in out pid)
    (process wish (list (string-append crunch-dir "/crunch-debugger.tcl")))
    (parameterize ((debugger-input out)
		   (debugger-active #t))
      (close-input-port in)
      (k)
      (display "complete_tags\n" (debugger-input))
      (close-output-port (debugger-input)))))

(define (register-pseudo-source-file fname txt)
  (register-source-file fname)
  (for-each
    (lambda (ln)
      (display (format "src ~s\n" ln) (debugger-input)))
    (string-split txt "\n")))

(define (register-source-file fname)
  (fprintf (debugger-input) "srcfile ~s\n" fname))

)

