;;; crunch - node data type
;
; nodes:
;
;   [unspecified]
;   [const (<datum> [<id>])]
;   [global-ref (<var> <exported?> <localname>)]
;   [local-ref (<alias> <var>)]
;   [if () <condition> <consequent> <alternative>]
;   [begin () <node> ...]
;   [global-set! (<var> <exported?> <localname>) <value>]
;   [local-set! (<alias> <var> <vartype>) <value>]
;   [let (<vars> <aliases>) <value> ... <body>]
;   [lambda (<id> <argc> <vars> <aliases> <where>) <body>]
;   [call (<tail?>) <function> <argument> ...]
;   [idcall (<id> <tail?> <opts?>) <argument> ...]
;   [dispatch (<idlist>) <function> <call> ...]
;   [variant (<id>) <lambda> ...]
;   [box-ref () <node>]
;   [box-set! () <box> <val>]
;   [labels (<labels>) <block> ...]
;   [goto (<label>)]
;   [c-external (<include> <name> (<argtype> ...) <rtype>)]
;   [c-lambda (<name> <rtype> ((<argtype> <param>) ...) (<string> ...))]
;   [c-declare ((<text> ...))]
;   [wrap (<wrapper>) <val>]
;   [switch ((<integer> ...)) <val> <node> ... <elsenode>]


(module (crunch compiler node) *

  (import (scheme base)
          (chicken base)
          (chicken plist)
          (chicken string)
          (chicken memory representation)
          matchable
          srfi-1)

  (define-record node
    location class data subnodes type)

  (define (node* loc class #!optional data . nodes)
    (make-node loc class data nodes #f))

  (define-record c-fragment data) ; used as placeholder during evaluation

  (define (clone-node n)
    (make-node (node-location n)
    	       (node-class n)
    	       (node-data n)
    	       (list-copy (node-subnodes n))
    	       (node-type n)))

  (define (replace-node! old new)	; retains loc and type
    (node-class-set! old (node-class new))
    (node-data-set! old (node-data new))
    (node-subnodes-set! old (node-subnodes new))
    old)

  (define (void-node! n)
    (node-class-set! n 'unspecified)
    (node-data-set! n '())
    (node-subnodes-set! n '())
    n)

  (define (ineffective? n)
    (case (node-class n)
      ((const unspecified local-ref global-ref box-ref) #t)
      ((lambda c-lambda)
        (let ((id (first (node-data n))))
          (and (null? (get id 'crunch:callers '()))
        	   (not (get id 'crunch:lambda-literal)))))
      ((begin if let variant labels wrap) (every ineffective? (node-subnodes n)))
      ((variant) (null? (node-subnodes n)))
      (else #f)))

  (define (nontrivial? n)
    (not (memq (node-class n) '(wrap const unspecified local-ref global-ref box-ref))))

  (define (bind-nontrivial ns gen k)
    (if (any nontrivial? ns)
        (let loop ((ns ns) (vars '()) (bs '()) (args '()))
          (cond ((null? ns)
                 (let ((body (k (reverse args))))
                   (make-node
                     (node-location body)
                     'let
                     (list vars vars)
                     (append bs (list body))
                     (node-type body))))
                ((nontrivial? (car ns))
                 (let ((var (gen)))
                  (loop (cdr ns)
                       (cons var vars)
                       (cons (car ns) bs)
                       (cons (make-node
                               (node-location (car ns))
                               'local-ref
                               (list var var)
                               '()
                               (node-type (car ns)))
                             args))))
                (else (loop (cdr ns) vars bs (cons (car ns) args)))))
        (k ns)))

  (define (drop-arg! n i)
    (assert (eq? 'idcall (node-class n)))
    (let loop ((i i) (args (node-subnodes n)) (new '()))
      (if (zero? i)
          (node-subnodes-set! n (append (reverse new) (cdr args)))
          (loop (sub1 i) (cdr args) (cons (car args) new)))))

  (define (localname a v)
    (if (eq? a v)
        v
        (string->symbol
          (string-append (symbol->string v) "/" (symbol->string a)))))

  (define (function-id id)
    (string->symbol (conc "<" id ">")))

  (define (quoted-expr n)
    (define (walk c)
      (cond ((procedure? c)
             (match (and (extended-procedure? c) (procedure-data c))
               (('crunch:lambda n)
                (function-id (first (node-data n))))
               (_ c)))
            ((symbol? c) (function-id c))
            ((pair? c) (cons (walk (car c)) (walk (cdr c))))
            ((vector? c)
              (list->vector (map walk (vector->list c))))
            (else c)))
    `(quote ,(walk (first (node-data n)))))

  (define (node-shape n #!optional (d 3) (decorator (lambda (_ x) x)))
    (define (sub d) (and d (sub1 d)))
    (let walk ((n n) (d d))
      (if (and d (zero? d))
          '...
          (decorator
            n
            (case (node-class n)
              ((const) (quoted-expr n))
              ((global-ref) (first (node-data n)))
              ((local-ref)
               (localname (first (node-data n)) (second (node-data n))))
              ((unspecified) '(%void))
              ((if begin)
               (let ((d2 (sub d)))
                 (cons (node-class n) (map (cut walk <> d2) (node-subnodes n)))))
              ((global-set!)
               `(set! ,(first (node-data n)) ,(walk (first (node-subnodes n)) (sub d))))
              ((local-set!)
               `(set! ,(localname (first (node-data n)) (second (node-data n)))
                  ,(walk (first (node-subnodes n)) (sub d))))
              ((let)
               `(let ,(map (lambda (v a n)
                             (list (localname a v) (walk n (sub d))))
                        (first (node-data n))
                        (second (node-data n))
                        (butlast (node-subnodes n)))
                  ,(walk (last (node-subnodes n)) (sub d))))
              ((lambda)
               (match (node-data n)
                 ((id argc vs as _)
                  `(%idlambda
                     ,(function-id id)
                     ,(let rec ((vs vs) (as as) (argc argc))
                        (cond ((eq? argc 'values)
                               (localname (car as) (car vs)))
                              ((null? vs) '())
                              ((eqv? 0 argc)
                               `(#!optional ,@(rec vs as #f)))
                              (else
                                (cons (localname (car as) (car vs))
                                      (rec (cdr vs) (cdr as) (and argc (sub1 argc)))))))
                     ,(if (get id 'crunch:integrated)
                          '<integrated>
                          (walk (first (node-subnodes n)) (sub d)))))))
              ((call)
               (let ((d2 (sub d)))
                 (map (cut walk <> d2) (node-subnodes n))))
              ((idcall)
               (let ((d2 (sub d)))
                 `(%idcall ,(function-id (first (node-data n)))
                           ,@(map (cut walk <> d2) (node-subnodes n)))))
              ((dispatch)
               (let* ((d2 (sub d))
                      (ns (node-subnodes n))
                      (calls (map (cut walk <> d2) (cdr ns))))
                 `(%dispatch ,(walk (car ns) d2) ,@calls)))
              ((wrap)
               `(%wrap ,(first (node-data n))
                       ,(walk (first (node-subnodes n)) (sub d))))
              ((variant)
               (let ((d2 (sub d)))
                 `(%variant ,@(map (cut walk <> d2) (node-subnodes n)))))
              ((box-ref) `(%box-ref ,(walk (first (node-subnodes n)) (sub d))))
              ((box-set!)
               (let ((d2 (sub d)))
                 `(%box-set! ,@(map (cut walk <> d2) (node-subnodes n)))))
              ((labels)
               (let ((d2 (sub d)))
                 `(%labels
                    ,@(append-map (lambda (l n)
                                    (list l (walk n d2)))
                        (node-data n)
                        (node-subnodes n)))))
              ((goto)
               `(%goto ,(first (node-data n))))
              ((switch)
               (let ((d2 (sub d))
                     (ns (node-subnodes n)))
                 `(%switch ,(walk (first ns) d2)
                           ,@(map (lambda (x n) (list x (walk n d2)))
                               (first (node-data n))
                               (butlast (cdr ns)))
                           ,(walk (last ns) d2))))
              ((c-lambda)
               `(%c-lambda ,@(node-data n)))
              ((c-external)
               `(%c-external ,@(node-data n)))
              ((c-declare)
               `(%c-declare ,@(node-data n)))
              (else => (cut error "bad node class" <>)))))))

  (define (node->expr n)
    (case (node-class n)
      ((const) (quoted-expr n))
      ((global-ref) (first (node-data n)))
      ((local-ref)
       (localname (first (node-data n)) (second (node-data n))))
      ((unspecified) '(%void))
      ((if begin)
       (cons (node-class n) (node-subnodes n)))
      ((global-set!)
       `(set! ,(first (node-data n)) ,(first (node-subnodes n))))
      ((local-set!)
       `(set! ,(localname (first (node-data n)) (second (node-data n)))
          ,(first (node-subnodes n))))
      ((let)
       `(let ,(map (lambda (v a n)
                     (list (localname a v) n))
                (first (node-data n))
                (second (node-data n))
                (butlast (node-subnodes n)))
          ,(last (node-subnodes n))))
      ((lambda)
       (match (node-data n)
         ((id argc vs as _)
          `(%idlambda
             ,(function-id id)
             ,(let rec ((vs vs) (as as) (argc argc))
                (cond ((null? vs) '())
                      ((eqv? 0 argc)
                       `(#!optional ,@(rec vs as #f)))
                      (else
                        (cons (localname (car as) (car vs))
                              (rec (cdr vs) (cdr as) (and argc (sub1 argc)))))))
             ,(if (get id 'crunch:integrated)
                  '<integrated>
                  (first (node-subnodes n)))))))
      ((call) (node-subnodes n))
      ((idcall)
       `(%idcall ,(function-id (first (node-data n)))
                 ,@(node-subnodes n)))
      ((dispatch)
       (let* ((ns (node-subnodes n))
              (calls (cdr ns)))
         `(%dispatch ,(car ns) ,@calls)))
      ((wrap)
       `(%wrap ,(first (node-data n))
               ,(first (node-subnodes n))))
      ((variant)
       `(%variant ,@(node-subnodes n)))
      ((box-ref) `(%box-ref ,(first (node-subnodes n))))
      ((box-set!)
       `(%box-set! ,@(node-subnodes n)))
      ((labels)
       `(%labels
          ,@(append-map list
              (node-data n)
              (node-subnodes n))))
      ((goto)
       `(%goto ,(first (node-data n))))
      ((switch)
       (let ((ns (node-subnodes n)))
         `(%switch ,(first ns)
                   ,@(map list
                       (first (node-data n))
                       (butlast (cdr ns)))
                   ,(last ns))))
      ((c-lambda)
       `(%c-lambda ,@(node-data n)))
      ((c-external)
       `(%c-external ,@(node-data n)))
      ((c-declare)
       `(%c-declare ,@(node-data n)))
      (else => (cut error "bad node class" <>))))

  (define (evaluate n)
    (define (void-node) (make-node (node-location n) 'unspecified '() '() #f))
    ;; toplevel expression
    (define (top n e)
      (case (node-class n)
        ((if)
          (match (node-subnodes n)
            ((t c a)
              (if (eval t e)
              	  (top c e)
              	  (top a e)))))
        ((begin)
          (let loop ((ns (node-subnodes n)) (ns2 '()))
            (match ns
              (() (void-node))
              ((n1) (cond ((null? ns2) (top n1 e))
              		     (else
              		       (node-subnodes-set! n (reverse (cons (top n1 e) ns2)))
              		       n)))
              ((n1 ns ...)
                (loop ns (cons (top n1 e) ns2))))))
        ((global-set!)
          (let* ((ns (node-subnodes n))
                 (body (eval* (first ns) e)))
            (set-car! ns body)
            n))
        ((local-set!)
          (let* ((v (second (node-data n)))
          	     (r (eval* (first (node-subnodes n)) e)))
          	(set-cdr! (assq v e) r)
		    (void-node)))
        ((let)
          (let* ((vs (first (node-data n)))
                 (vals (map (cut eval* <> e) (butlast (node-subnodes n))))
                 (body (last (node-subnodes n)))
                 (e2 (append
                       (map (lambda (var val)
                       	      (if (eq? 'const (node-class val))
                       	     	  (cons var (first (node-data val)))
                       	   	      (cons var (eval val e))))
                       	   vs vals)
                       e))
                 (bs (filter-map
                       (lambda (var alias val)
                         (and (not (eq? 'const (node-class val)))
                         	  (list var alias val)))
                       vs
                       (second (node-data n))
                       (butlast (node-subnodes n)))))
            (cond ((null? bs) (top body e2))
            	  (else
            	    (node-data-set! n (list (map first bs) (map second bs)))
            	    (node-subnodes-set! n (append (map third bs) (list (top body e2))))
            	    n))))
        ((unspecified lambda) (void-node))
        (else n)))
    ;; non-toplevel, inner expression
    (define (eval* n e)
      (case (node-class n)
        ((if)
          (match (node-subnodes n)
            ((t c a)
              (if (eval t e)
              	  (eval* c e)
              	  (eval* a e)))))
        ((begin)
          (let loop ((ns (node-subnodes n)) (ns2 '()))
            (match ns
              (() (void-node))
              ((n1) (cond ((null? ns2) (eval* n1 e))
              		     (else
              		       (node-subnodes-set! n (reverse (cons (eval* n1 e) ns2)))
              		       n)))
              ((n1 ns ...)
                (loop ns (cons (eval* n1 e) ns2))))))
        ((let)
          (let* ((vs (first (node-data n)))
                 (vals (map (cut eval* <> e) (butlast (node-subnodes n))))
                 (body (last (node-subnodes n)))
                 (e2 (append
                       (map (lambda (var val)
                       	      (if (eq? 'const (node-class val))
                       	     	  (cons var (first (node-data val)))
                       	   	      (cons var (eval val e))))
                       	   vs vals)
                       e)))
            (node-subnodes-set! n (append vals (list (eval* body e2))))
            n))
        ((lambda c-lambda c-external c-declare) n)
        (else (make-node (node-location n) 'const (list (eval n e)) '() #f))))
    ;; evaluated expression
    (define (walk n)
      (case (node-class n)
        ((unspecified) (lambda (e) (void)))
        ((const)
          (let ((x (first (node-data n))))
            (lambda (e) x)))
        ((global-ref)
          (let ((var (first (node-data n))))
            (lambda (e)
              (if (##sys#symbol-has-toplevel-binding? var)
                  (##sys#slot var 0)
                  (error "unbound variable" var)))))
        ((local-ref)
          (let ((v (second (node-data n))))
            (lambda (e) (cdr (assq v e)))))
        ((if)
          (match (node-subnodes n)
            ((t c a)
              (let ((t (walk t))
                    (c (walk c))
                    (a (walk a)))
                (lambda (e) (if (t e) (c e) (a e)))))))
        ((begin)
          (let ((ns (map walk (node-subnodes n))))
            (lambda (e)
              (let loop ((ns ns))
                (if (null? (cdr ns))
                    ((car ns) e)
                    (begin
                      ((car ns) e)
                      (loop (cdr ns))))))))
        ((global-set!)
          (let ((var (first (node-data n)))
                (val (walk (first (node-subnodes n)))))
            (lambda (e)
              (##sys#setslot var 0 (val e))
              (void))))
        ((local-set!)
          (let ((v (second (node-data n)))
                (val (walk (first (node-subnodes n)))))
            (lambda (e)
              (set-cdr! (assq v e) (val e))
              (void))))
        ((let)
          (let ((vs (first (node-data n)))
                (vals (map walk (butlast (node-subnodes n))))
                (body (walk (last (node-subnodes n)))))
            (lambda (e)
              (body (append (map (lambda (a x) (cons a (x e))) vs vals) e)))))
        ((lambda)
          (match (node-data n)
            ((_ argc vs _ _)
             (let* ((lbody (first (node-subnodes n)))
                    (body (walk lbody)))
               (lambda (e)
                 (extend-procedure
                   (lambda args
                     (let loop ((i 0) (args args) (as vs) (e e))
                       (cond ((null? as) (body e))
                             ((>= i argc)
                              (if (null? args)
                                  (loop i '() (cdr as) (cons (cons (car as) #f) e))
                                  (loop i (cdr args) (cdr as) (cons (cons (car as) (car args)) e))))
                             (else
                               (loop (add1 i) (cdr args) (cdr as)
                                     (cons (cons (car as) (car args)) e))))))
                   `(crunch:lambda ,n)))))))
        ((call)
          (match (node-subnodes n)
            ((fn . args)
              (let ((fn (walk fn))
                    (args (map walk args)))
                (lambda (e)
                  (apply (fn e) (map (lambda (x) (x e)) args)))))))
        ((c-lambda c-external c-declare)
          (let ((x (make-c-fragment (node-data n))))
            (lambda (e) x)))
        ;; idcall, dispatch, variant, box, box-ref, box-set!, goto, labels, wrap, switch
        ;; are compiler-specific
        (else => (cut error "invalid node class while evaluating node" <>))))
    (define (eval n e) ((walk n) e))
    (top n '()))

  (define (cleanup-node n)
    (case (node-class n)
      ((local-ref global-ref local-set! global-set!)
        (let ((var (first (node-data n))))
          (unless (get var 'crunch:primitive)
            (remprop! var 'crunch:function-id)
            (remprop! var 'crunch:type))
          (drop-props
            var
            ;; not: crunch:inline, crunch:primitive
            '(crunch:used crunch:assigned crunch:free-in crunch:owner crunch:declared-type
                          crunch:globalized crunch:type* crunch:boxed
                          crunch:freevar-type crunch:initial-value crunch:ambiguous))))
      ((label)
        (for-each (cut drop-props <> '(crunch:seen))
          (first (node-data n))))
      ((lambda)
        (let ((id (first (node-data n))))
          (for-each cleanup-node
            (append
              (get id 'crunch:variant '()))
              (get id 'crunch:call-sites '()))
          (drop-props
            id
            ;; not: crunch:primitive
            '(crunch:freevars crunch:addfree crunch:callers crunch:lambda-node
              crunch:escapes crunch:call-sites crunch:lambda crunch:overload
              crunch:external-name crunch:specialize crunch:variant crunch:escapes
              crunch:seen crunch:extern crunch:integrated crunch:variant-of
              crunch:instantiated crunch:lambda-literal)))))
    (for-each cleanup-node (node-subnodes n)))

    (define (drop-props sym lst)
      (for-each (cut remprop! sym <>) lst))

)
