;;;; 6502 emulator

(module mos6502 (execute!
                 reset!
                 interrupt!
                 make-processor
                 processor-memory
                 processor-registers
                 processor?)

  (import (scheme base)
          (scheme write)
          (chicken base)
  	      (chicken bitwise)
  	      (chicken number-vector)
  	      (chicken fixnum)
  	      (chicken foreign)
  	      (chicken format)
  	      crunch
  	      (except miscmacros begin0))

  (define-syntax-rule (begin0 x more ...)
    (let ((r x)) (begin more ...) r))

  (define-record-type processor
    (%make-processor memory registers)
    processor?
    (memory processor-memory)
    (registers processor-registers)) ; #u(A X Y P S PC(lo) PC(hi))

  (define (make-processor #!key memory registers)
    (%make-processor
      (or memory (make-u8vector #x10000 #xff))
      (or registers (u8vector 0 0 0 #x20 0 0 0))))

  (define-syntax-rule (hi n) (fxshr n 8))
  (define-syntax-rule (lo n) (fxand #xff n))
  (define-syntax-rule (lohi n m) (fxior n (fxshl m 8)))
  (define-syntax-rule (word n) (fxand #xffff n))
  (define-syntax-rule (byte n) (fxand #xff n))

  (define-syntax-rule (peek M a)
    (u8vector-ref M (word a)))

  (define-syntax-rule (poke! M a n)
    (u8vector-set! M (word a) (byte n)))

  (define-syntax-rule (peekw M a)
    (let ((a1 a))
      (lohi (peek M a1) (peek M (+ a1 1)))))

  (define-syntax-rule (pokew! M a n)
    (let ((a1 a)
          (n1 n))
      (poke! M a1 (lo n1))
      (poke! M (+ a1 1) (hi n1))))

  (define-syntax-rule (or! v n)
    (set! v (fxior v n)))

  (define-syntax-rule (and! v n)
    (set! v (fxand v n)))

  (define-syntax-rule (xor! v n)
    (set! v (fxxor v n)))

  (define-syntax incb!
    (syntax-rules ()
      ((_ n m) (set! n (byte (+ n m))))
      ((_ n) (incb! n 1))))

  (define-syntax incw!
    (syntax-rules ()
      ((_ n m) (set! n (word (+ n m))))
      ((_ n) (incw! n 1))))

  (define-syntax decb!
    (syntax-rules ()
      ((_ n m) (set! n (byte (- n m))))
      ((_ n) (decb! n 1))))

  (define-syntax decw!
    (syntax-rules ()
      ((_ n m) (set! n (word (- n m))))
      ((_ n) (decw! n 1))))

  (define-syntax-rule (nflag? n)
    (not (zero? (fxand #x80 n))))

  (define-syntax-rule (zflag! P n)
    (let ((n1 (byte n)))
      (if (zero? n1)
          (or! P #x02)
          (and! P #xfd))))

  (define-syntax-rule (nflag! P n)
    (let ((n1 n))
      (if (zero? (fxand n1 #x80))
          (and! P #x7f)
          (or! P #x80))))

  (define-syntax-rule (sextend n)
    (fx- (fxxor (byte n) #x80) #x80))

  (cond-expand
    (simulate (include "engine.scm"))
    (else
      (crunch-options verbose)
      (crunch
        (: (engine integer (numvector u8) (numvector u8)) integer)
        (include "engine.scm"))))

  (define (reset! p)
    (let ((regs (processor-registers p))
          (mem (processor-memory p)))
      (u8vector-set! regs 0 0)          ; A
      (u8vector-set! regs 1 0)          ; X
      (u8vector-set! regs 2 0)          ; Y
      (u8vector-set! regs 3 #x20)       ; P
      (u8vector-set! regs 4 #xff)       ; S
      (u8vector-set! regs 5 (u8vector-ref mem #xfffc)) ; PC(lo)
      (u8vector-set! regs 6 (u8vector-ref mem #xfffd)))) ; PC(hi)

  (define (interrupt! p vec)
    (let* ((regs (processor-registers p))
           (mem (processor-memory p))
           (S (u8vector-ref regs 4))
           (PC (fx+ 1 (fxior (u8vector-ref regs 5)
                             (fxshl (u8vector-ref regs 6) 8)))))
      (u8vector-set! mem (fx+ #x100 S) (fxshr PC 8))
      (u8vector-set! mem (fx+ #xff S) (fxand #xff PC))
      (u8vector-set! mem (fx+ #xfe S) (u8vector-ref regs 3)) ; push P
      (u8vector-set! regs 3 (fxior (u8vector-ref regs 3) #x04)) ; set I flag
      (u8vector-set! regs 4 (fxand #xff (fx- S 3)))
      (u8vector-set! regs 5 (u8vector-ref mem vec)) ; PC(lo)
      (u8vector-set! regs 6 (u8vector-ref mem (fxand #xff (fx+ vec 1)))))) ; PC(hi)

  (define (execute! p #!key cycles)
    (let ((M (processor-memory p))
          (regs (processor-registers p)))
      (let loop ()
        (let ((c2 (engine (or cycles 999999) M regs)))
          (if cycles
              c2
              (loop))))))

)
